# -*- coding: utf-8 -*-

import os
import sys
import re
import argparse
import json
import hashlib
import shutil
import pickle
from collections import namedtuple, defaultdict

from .packages import gnupg
from .packages import argh
from .packages.argh import arg, named

__version__ = '97'

verbose = False
always_rebuild = False

cache = {}
gpg = None

vit_dir = '.vit'
cache_file = os.path.join(vit_dir, 'cache')
config_file = os.path.join(vit_dir, 'config')
gpg_dir = os.path.join(vit_dir, 'gnupg')

wot_dir = 'wot'
sig_dir = 'sigs'
patch_dir = 'patches'
req_dirs = [vit_dir, wot_dir, sig_dir, patch_dir]


def fatal(msg):
    sys.stderr.write(msg + "\n")
    exit(1)

def spew(msg):
    if verbose:
        print msg

# List of files in a directory, in lexical order.
def dir_files(d):
    return sorted(next(os.walk(d))[2])

def sha512(filepath):
    with open(filepath, 'r') as f:
        h = hashlib.sha512(f.read())
        return h.hexdigest()

Filehash = namedtuple('Filehash',['h','p'])

##############################################################################

def toposort(unsorted):
    sorted = []
    unsorted = dict(unsorted)
    while unsorted:
        acyclic = False
        for node, edges in unsorted.items():
            for edge in edges:
                if edge in unsorted:
                    break
            else:
                acyclic = True
                del unsorted[node]
                sorted.append((node, edges))
        if not acyclic:
            fatal("Cyclic graph!")
    return sorted

# Grep for diff magics
def vpdata(content, exp):
    l = []
    for m in re.findall(exp, content, re.MULTILINE):
        path = m[0].split('/', 1)[1] # remove a/,b/,etc prefix
        h = m[1]
        l.append(Filehash(h,path))
    return l

# Parent files of a vpatch
def vp_parents(content):
   return vpdata(content, r'^--- (\S+) (\S+)$')

# Child files of a vpatch
def vp_children(content):
    return vpdata(content, r'^\+\+\+ (\S+) (\S+)$')


# It is entirely possible to have more than one root!
# ... exactly how, is left as an exercise for readers.
def find_roots():
    patches = cache['patches']
    roots = []
    # Walk, find roots
    for pf in patches:
        if all(p.h == 'false' for p in patches[pf]['parents']):
            roots.append(pf)
    return roots

# Get antecedents.
def get_ante(name):
    patch = cache['patches'][name]
    ante = defaultdict(list)
    for parent in patch['parents']:
        if parent.h == 'false':
            continue
        origin = cache['origin'][parent]
        ante[origin].append(parent)
    return ante

# Get descendants.
def get_desc(name):
    patch = cache['patches'][name]
    desc = defaultdict(list)
    for child in patch['children']:
        if child.h == 'false' or child not in cache['offspring']:
            continue
        for p in cache['offspring'][child]:
            desc[p].append(child)
    return desc

# Recursively get descendants
def re_desc(name):
    stack = get_desc(name).keys()
    desc = set(stack)
    while stack:
        p = stack.pop()
        l = get_desc(p).keys()
        for d in l:
            if d not in desc:
                stack.append(d)
                desc.add(d)
    return desc
    
def find_conflicts():
    conflicts = {}
    for h,children in cache['offspring'].iteritems():
        if len(children) > 1:
            parent = cache['origin'][h]
            if not parent in conflicts:
                conflicts[parent] = {}
            conflicts[parent][h] = children
    return conflicts

def deconflict(patchset):
    conflicts = find_conflicts()
    if not conflicts:
        return patchset

    newpatchset = list(patchset)
    removed = set()

    for pf in patchset:
        if pf in conflicts and pf not in removed:
            for h,children in conflicts[pf].iteritems():
                # skip conflicts we've already removed
                branches = set(children) - removed
                # skip conflicts not in patchset
                branches &= set(patchset)
                if len(branches) > 1:
                    choice = choose(branches)
                    branches.remove(choice)
                    for branch in branches:
                        removed.add(branch)
                        removed |= re_desc(branch)

    for r in removed:
        if r in newpatchset:
            newpatchset.remove(r)

    print 'Removed patches:'
    for r in sorted(removed, key=patchset.index):
        print '  {0}'.format(r)
    return newpatchset

def choose(choices, title='Choose a branch:'):
    choices = list(choices)
    print '{0}: '.format(title)
    for i,v in enumerate(choices):
        print '  [{0}]:  {1}'.format(i,v)
    i = raw_input('Enter number: ') 
    return choices[int(i)]

def patch_name(s):
    if '/' in s:
        s = os.path.basename(s)
    if s not in cache['patches']:
        fatal('Patch not found: {0}'.format(s))
    return s

# Print name of patch and its guarantors, or 'UNSIGNED' if none known.
def disp_vp(name):
    sigs = cache['patches'][name]['sigs']
    pubkeys = cache['pubkeys']
    if not sigs:
        s = 'UNSIGNED'
    else:
        s = ','.join(pubkeys[fp]['handle'] for fp in sigs)
            
    return "{0}   ({1})".format(name, s)

########################################################################

def scan_files():
    files = []
    for dr in [wot_dir,patch_dir,sig_dir]:
        dirfiles = []
        for f in dir_files(dr):
            fp = os.path.join(dr, f)
            h = sha512(fp)
            dirfiles.append(Filehash(h, f))
        files.append(dirfiles)
    return tuple(files)

def save_cache(data):
    with open(cache_file, 'w') as f:
        pickle.dump(data, f)

def load_cache():
    if not os.path.isfile(cache_file):
        return dict()
    with open(cache_file, 'r') as f:
        return pickle.load(f)

def files_changed(old, new):
    for i in [0,1,2]:
        if not (set(old[i]) == set(new[i])):
            return True
    return False

def build_wot(keyfiles):
    pubkeys = {}
    handle = {}
    for filehash in keyfiles:
        pk = filehash.p
        with open(os.path.join(wot_dir, pk), 'r') as f:
            pubkey = f.read()
        impkey = gpg.import_keys(pubkey)
        for fp in impkey.fingerprints:
            handle[fp] = os.path.splitext(pk)[0]

    for k in gpg.list_keys():
        name = handle[k['fingerprint']]
        pubkeys[k['keyid']] = {'fp':k['fingerprint'],
                               'id':', '.join(map(str, k['uids'])),
                               'handle':name}
    return pubkeys

def process_patches(pfiles, sfiles):
    patches = {}
    unsigned = []
    origin = {}
    offspring = defaultdict(list)
    
    for filehash in pfiles:
        pf = filehash.p
        pdata = {'sigs':[]}

        ppath = os.path.join(patch_dir, pf)
        with open(ppath, 'r') as f:
            patch = f.read()
        pdata['h'] = hashlib.sha512(patch).hexdigest()

        parents = vp_parents(patch)
        children = vp_children(patch)

        if not children or not parents:
             fatal('{0} is INVALID, check whether it IS a vpatch!'.format(pf))

        for child in children:
            if child.h != 'false': # false means file was deleted
                origin[child] = pf
                
        for parent in parents:
            if parent.h != 'false': # false means file was created
                offspring[parent].append(pf)
            
        pdata['parents'] = parents
        pdata['children'] = children

        for filehash in sfiles:
            sf = filehash.p
            spath = os.path.join(sig_dir, sf)
            # All seals must take the form patchtitle.vpatch.yourname.sig
            if sf.find(pf) == 0: # substring of sig filename up through '.vpatch'
                v = gpg.verify_file(patch, spath)
                if v.valid:
                   pdata['sigs'].append(v.key_id)
                else:
                    fatal("---------------------------------------------------------------------\n" +
                          "WARNING: {0} is an INVALID seal for {1} !\n".format(sf, pf) +
                          "Check that this user is in your WoT, and that this key has not expired.\n" +
                          "Otherwise remove the invalid seal from your SEALS directory.\n" +
                          "---------------------------------------------------------------------")

        patches[pf] = pdata

    return (patches, origin, offspring)

def rebuild(files):
    global cache, gpg
    cache = {}
    
    if os.path.isdir(gpg_dir):
        shutil.rmtree(gpg_dir)
    gpg = gnupg.GPG(homedir=gpg_dir)
    
    keyfiles, pfiles, sfiles = files
    pubkeys = build_wot(keyfiles)
    patches, origin, offspring = process_patches(pfiles, sfiles)

    unsigned = [p for p in patches if not patches[p]['sigs']]

    cache.update({
        'patches': patches,
        'pubkeys': pubkeys,
        'unsigned': unsigned,
        'origin': origin,
        'offspring': offspring,
        'files': files,
        })

    roots = find_roots()
    if not roots:
        fatal('No roots found!')

    # Topological ordering of flow graph
    t = []
    for pf in patches:
        t.append((pf, get_desc(pf).keys()))
    toposorted = [x[0] for x in toposort(t)]
    toposorted.reverse()

    cache.update({
        'roots': roots,
        'toposorted': toposorted,
        })

def prepare_cache(args):
    global cache, always_rebuild

    if args.get_function() == vit_init:
        return # don't need cache for init command
    elif args.get_function() == vit_rebuild:
        always_rebuild = True

    for d in req_dirs:
        if not os.path.isdir(d):
            fatal('{} directory not found, try `vit init`'.format(d))
    
    cache = load_cache()
    files = scan_files()
    if not cache or always_rebuild or files_changed(cache['files'], files):
        rebuild(files)
        save_cache(cache)


def main():
    parser = argh.ArghParser()
    cmds = [vit_init, vit_rebuild, vit_json, wot, flow, roots, ante, desc, press, seq, conflicts, origin, info]
    parser.add_commands(cmds)
    parser.dispatch(pre_call=prepare_cache)


## Commands
##############################################################################


@named('init')
def vit_init():
    """Initialize needed dirs"""
    if os.path.isdir(vit_dir):
        fatal('.vit already exists in current dir')
    for d in req_dirs:
        if not os.path.isdir(d):
            os.mkdir(d)
            print 'Created ./{0}'.format(d)


@named('rebuild')
def vit_rebuild():
    """Force rebuild of vit cache"""
    print 'Cache was rebuilt.'


def wot():
    """Show trusted names and key fingerprints"""
    for k in cache['pubkeys'].values():
        print '{0} :: {1}'.format(k['handle'], k['fp'])
        print '  ({0})'.format(k['id'])

@arg('-w', '--wild', help='Allow unsigned patches')
def flow(wild=False):
    """Show toposorted patch flow"""
    seq = cache['toposorted']
    if not wild:
        seq = [p for p in seq if p not in cache['unsigned']]
    for p in seq:
        print disp_vp(p)


def roots():
    """Show root patches"""
    for r in cache['roots']:
        print 'Root: {0}'.format(disp_vp(r))


def print_relatives(vpatch, rel):
    func = get_ante if rel == 'ante' else get_desc
    dic = func(patch_name(vpatch))
    if not dic:
        print 'None found.'
    keys = dic.keys()
    keys.sort(key=cache['toposorted'].index)
    for p in keys:
        print '  {0}'.format(disp_vp(p))
        for f in dic[p]:
            print '    {0}'.format(f.p)

def ante(vpatch):
    """Find antecedents of vpatch"""
    print_relatives(vpatch, 'ante')

def desc(vpatch):
    """Find descendants of vpatch"""
    print_relatives(vpatch, 'desc')


def press_patches(sequence, dest):
    os.mkdir(dest)
    for pf in sequence:
        print 'Using: {0}'.format(disp_vp(pf))
        ppath = os.path.join(patch_dir, pf)
        os.system("patch -E --dir {0} -p1 < {1}".format(dest, ppath))
    print "Completed Pressing using head: {0} to path: '{1}'".format(sequence[-1], dest)


@arg('-w', '--wild', help='Allow unsigned patches')
@arg('-s', '--seq', help='Export sequence file instead of source tree')
def press(vpatch, dest, wild=False, seq=False):
    """Create a source tree"""
    head = patch_name(vpatch)
    headpos = cache['toposorted'].index(head)
    sequence = cache['toposorted'][:headpos + 1]
    if not wild:
        sequence = [p for p in sequence if p not in cache['unsigned']]
    sequence = deconflict(sequence)

    if seq:
        with open(dest, 'w') as f:
            lines = ['{0}\t{1}'.format(pf, cache['patches'][pf]['h']) for pf in sequence]
            f.write('\n'.join(lines))
        return

    print "Pressing using head: {0} to path: '{1}'".format(sequence[-1], dest)
    press_patches(sequence, dest)


def seq(seqfile, dest):
    """Press a sequence of patches"""
    with open(seqfile, 'r') as f:
        lines = f.readlines()
    sequence = [p.split('\t')[0] for p in lines]
    press_patches(sequence, dest)


def conflicts():
    """Find conflicts"""
    confs = find_conflicts()
    if not confs:
        print 'No conflicts found.'
    for pf, files in confs.iteritems():
        print 'Origin: {0}'.format(disp_vp(pf))
        for f, children in files.iteritems():
            print 'Conflict on: {0}'.format(f.p)
            for c in children:
                print '  {0}'.format(c)
            print ' '

@arg('filepath', help='Path to file or sha512 hash')
def origin(filepath):
    """Find origin patch for file"""
    if os.path.isfile(filepath):
        s = sha512(filepath)
    else:
        s = filepath

    for f in cache['origin']:
        if f.h == s:
            print 'Origin: {0}'.format(disp_vp(cache['origin'][f]))
            print 'Filename: {0}'.format(f.p)
            print 'Hash: {0}'.format(f.h)
            print 'Offspring:'
            if f in cache['offspring']:
                for p in cache['offspring'][f]:
                    print '  {0}'.format(p)
            return
    print 'No origin known.'


def info(vpatch):
    """Show information about vpatch"""
    try:
        from unidiff import PatchSet
    except ImportError:
        fatal('This command requires the unidiff package. Try `pip install unidiff`.')

    pf = patch_name(vpatch)
    ppath = os.path.join(patch_dir, pf)
    patch = PatchSet.from_filename(ppath)

    print 'Patch: {0}'.format(disp_vp(pf))
    print 'Hash: {0}'.format(cache['patches'][pf]['h'])
    print '\n=== Modifications ===\n'
    additions = 0
    deletions = 0
    for f in patch:
        additions += f.added
        deletions += f.removed
        print '  {0}: +{1} additions, -{2} deletions'.format(f.path.split(' ',1)[0], f.added, f.removed)
    print ' '
    print ' {0} modified file(s), {1} added file(s), {2} removed file(s)'.format(
        len(patch.modified_files),
        len(patch.added_files),
        len(patch.removed_files)
        )
    print ' Total: {0} addition(s), {1} deletion(s)'.format(additions, deletions)
    print '\n=== Descendants ===\n'
    desc(pf)
    print '\n=== Antecedents ===\n'
    ante(pf)


@named('json')
@arg('dest', nargs='?', default='vit.json', help='Destination filepath')
def vit_json(dest):
    """Export vit cache to json file"""
    cache['origin'] = {'{0} {1}'.format(k.h, k.p):v for k,v in cache['origin'].iteritems()}
    cache['offspring'] = {'{0} {1}'.format(k.h, k.p):v for k,v in cache['offspring'].iteritems()}
    with open(dest, 'w') as f:
        json.dump(cache, f)
        print 'Exported json to {0}'.format(dest)
